/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 * http://www.jcuda.org
 *
 * Copyright 2009-2012 Marco Hutter - http://www.jcuda.org
 */

package org.jcuda.samples.matrix.cg;

import static jcuda.jcusparse.JCusparse.cusparseDestroyMatDescr;
import static jcuda.runtime.JCuda.cudaFree;
import jcuda.driver.CUdeviceptr;
import jcuda.jcusparse.cusparseMatDescr;

/**
 * Default implementation of a {@link FloatMatrixDeviceCSR}
 */
class DefaultFloatMatrixDeviceCSR implements FloatMatrixDeviceCSR
{
    /**
     * The CUSPARSE matrix descriptor
     */
    private final cusparseMatDescr descr;
    
    /**
     * The number of rows 
     */
    private final int numRows;
    
    /**
     * The number of columns 
     */
    private final int numCols;

    /**
     * The number of non-zero elements in this matrix
     */
    private final int numNonZeros;
    
    /**
     * The non-zero values 
     */
    private final CUdeviceptr values;
    
    /**
     * The row pointers
     */
    private final CUdeviceptr rowPointers;
    
    /**
     * The column indices 
     */
    private final CUdeviceptr columnIndices;
    
    /**
     * Creates a new matrix with the given values. All the given
     * values will be stored directly. Thus, freeing the given
     * pointers will be done in the {@link #dispose()} method
     * of this class.
     * 
     * @param descr The CUSPARSE matrix descriptor
     * @param numRows The number of rows 
     * @param numCols The number of columns
     * @param numNonZeros The number of non-zero elements 
     * @param values The non-zero values
     * @param rowPointers The row pointers
     * @param columnIndices The column indices
     */
    DefaultFloatMatrixDeviceCSR(cusparseMatDescr descr, 
        int numRows, int numCols, int numNonZeros, 
        CUdeviceptr values, CUdeviceptr rowPointers, 
        CUdeviceptr columnIndices)
    {
        this.descr = descr;
        this.numRows = numRows;
        this.numCols = numCols;
        this.numNonZeros = numNonZeros;
        this.values = values;
        this.rowPointers = rowPointers;
        this.columnIndices = columnIndices;
    }
    
    @Override
    public void dispose()
    {
        cudaFree(values);
        cudaFree(rowPointers);
        cudaFree(columnIndices);
        cusparseDestroyMatDescr(descr);
    }
    
    @Override
    public int getNumRows()
    {
        return numRows;
    }

    @Override
    public int getNumCols()
    {
        return numCols;
    }
    
    @Override
    public int getNumNonZeros()
    {
        return numNonZeros;
    }

    @Override
    public CUdeviceptr getValues()
    {
        return values;
    }

    @Override
    public CUdeviceptr getRowPointers()
    {
        return rowPointers;
    }

    @Override
    public CUdeviceptr getColumnIndices()
    {
        return columnIndices;
    }
    
    @Override
    public cusparseMatDescr getDescriptor()
    {
        return descr;
    }
    

    @Override
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        sb.append("Copied from device:\n");
        sb.append(Matrices.createHostCSR(this));
        return sb.toString();
    }


}